// PDP-8
// Based on descriptions in "Computer Engineering"
// Nov 2005 Brad Parker brad@heeltoe.com

//  0  1  2   3   4   5   6   7   8   9  10  11
// 11 10  9   8   7   6   5   4   3   2   1   0
// |--op--|
// 0 and
// 1 tad
// 2 isz
// 3 dca
// 4 jms
// 5 jmp
// 6 iot
// 7 opr
// 11 10  9   8   7   6   5   4   3   2   1   0
// group 1
//            x
//               |cla|clf|       |          |
//               |   |   |cma cml|          |
//                               |rar       |
//                               |ral       |
//                               |rtr       |
//                               |rrl       |
//                               |          |iac
//
// 11 10  9   8   7   6   5   4   3   2   1   0
// group 2
//            x
//                   |sma|sza|snl|skp|      |
//               |cla|
//                                   |osr|hlt
//
// 11 10  9   8   7   6   5   4   3   2   1   0
// eae
//            x
//               |cla|
//                   |mqa|sca|mql|
//                               |isn
// 

/*
 cpu states
 
  F0 fetch
  F1 incr pc
  F2 write
  F3 dispatch

  E0 read
  E1 decode
  E2 write
  E3 load

or

  D0 read
  D1 wait
  D2 write
  D3 load

 ------
 
  F0 fetch
	check for interrupt
  F1 incr pc
	if opr
		group 1 processing
 		group 2 processing
 
 	if iot
 
	incr pc or skip (incr pc by 2)
 
  F2 write
 	ma <= pc
 
  F3 dispatch
	if opr
		group1 processing
 
	if !opr && !iot
 		possible defer
 
 
  D0
	mb <= memory
  D1
*/

module pdp8(clk, reset_n, switches);

input clk, reset_n;
input[11:0] switches;

// mmeory buffer, holds data, instructions
reg[11:0] mb;

// hold address of work in memory being accessed
reg[11:0] ma;

reg[11:0] ac;
reg l;

reg[11:0] pc;
wire pc_incr, pc_skip;

reg[2:0] ir;

//
wire[11:0] memory_bus;
reg ram_we_n;

//
reg[3:0] state, next_state;

reg run;
reg interrupt_enable;
reg interrupt;
reg interrupt_cycle;

reg io_pulse_1, io_pulse_2, io_pulse_4;

reg rx_int, tx_int;
reg[12:0] rx_data, tx_data;
reg tx_delaying;
reg[3:0] tx_delay;

wire[5:0] io_select;

assign io_select = mb[8:3];

wire skip_condition;

wire fetch;	// memory cycle to fetch instruction
wire deferred;	// memory cycle to get address of operand
wire execute;	// memory cycle to getch (store) operand and execute isn

assign {fetch, deferred, execute} =
	(state[3:2] == 2'b00) ? 3'b100 :
	(state[3:2] == 2'b01) ? 3'b010 :
	(state[3:2] == 2'b10) ? 3'b001 :
				3'b000 ;

wire and,tad,isz,dca,jms,jmp,iot,opr;

assign {and,tad,isz,dca,jms,jmp,iot,opr} =
	(ir == 3'b000) ? 8'b10000000 :
	(ir == 3'b001) ? 8'b01000000 :
	(ir == 3'b010) ? 8'b00100000 :
	(ir == 3'b011) ? 8'b00010000 :
	(ir == 3'b100) ? 8'b00001000 :
	(ir == 3'b101) ? 8'b00000100 :
	(ir == 3'b110) ? 8'b00000010 :
                         8'b00000001 ;


//-------------

parameter F0 = 4'b0000;
parameter F1 = 4'b0001;
parameter F2 = 4'b0010;
parameter F3 = 4'b0011;

parameter D0 = 4'b0100;
parameter D1 = 4'b0101;
parameter D2 = 4'b0110;
parameter D3 = 4'b0111;

parameter E0 = 4'b1000;
parameter E1 = 4'b1001;
parameter E2 = 4'b1010;
parameter E3 = 4'b1011;

ram_4kx12 ram(
	.A(ma),
	.DI(mb),
	.DO(memory_bus),
	.CE_N(1'b0),
	.WE_N(ram_we_n));

assign skip_condition =
		(mb[6] && ac[11]) ||
		(mb[5] && ac == 0) ||
		(mb[4] && l == 0);

assign pc_incr =
	(opr & !mb[8]) ||
	(opr && (mb[8] && !mb[0]) && (skip_condition == mb[3])) ||
	iot ||
	(!(opr || iot) && !interrupt_cycle);

assign pc_skip =
	(opr && (mb[8] && !mb[0]) && (skip_condition ^ mb[3])) ||
	(iot && mb[0] && io_select == 6'o03 && rx_int) ||
	(iot && mb[0] && io_select == 6'o04 && tx_int);

always @(reset_n)
  if (reset_n == 0)
    begin
	pc <= 0;
        state <= 0;
        next_state <= 0;
	ma <= 0;
	mb <= 0;
	ac <= 0;
	l <= 0;
	ir <= 0;
	ram_we_n <= 1;
	run <= 1;
	interrupt_enable <= 0;
	interrupt <= 0;
	io_pulse_1 <= 0;
	io_pulse_2 <= 0;
	io_pulse_4 <= 0;
    end

initial
  begin
    ram_we_n = 1;
  end

/*
 * cpu state state machine
 * 
 * clock next cpu state at rising edge of clock
 */

always @(posedge clk)
  state <= #1 next_state;

always @(state)
  begin
    case (state)

      // FETCH 
      F0:
	begin
	  if (tx_delaying)
	    begin
	      tx_delay <= tx_delay - 1;
	      if (tx_delay == 4'b0)
		begin
		  tx_delaying <= 0;
		  interrupt <= 1;
		end
	    end

	  if (interrupt && interrupt_enable && !interrupt_cycle)
	    begin
$display("interrupt; %b %b %b", interrupt, interrupt_enable, interrupt_cycle);
	      interrupt_cycle <= 1;
	      interrupt <= 0;
	      mb <= 12'o4000;
	      ir <= 3'o4;
	    end
	  else
	    begin
	      interrupt_cycle <= 0;
	      mb <= memory_bus;
	      ir <= memory_bus[11:9];
	    end

	  next_state <= F1;
	end

      F1:
	begin
//$display("f1 - ma %o, mb %o, ir %o", ma, mb, ir);
	if (opr)
	  begin
		// group 1
		if (!mb[8])
		  begin
//		    pc <= pc + 1;
		    if (mb[7]) ac <= 0;
		    if (mb[6]) l <= 0;
		    if (mb[5]) ac <= ~ac;
		    if (mb[4]) l <= ~l;
		  end

		// group 2
		if (mb[8] & !mb[0])
		  begin
//		    if (skip_condition ^ mb[3])
//		      pc <= pc + 2;
//		    if (skip_condition == mb[3])
//		      pc <= pc + 1;
//		    else
		      if (mb[7])
		        ac <= 0;
		  end
	  end

	if (iot)
	  begin
$display("iot %t, run %b, state %b, pc %o, ir %o, mb %o, io_select %o",
	$time, run, state, pc, ir, mb, io_select);
//		pc <= pc + 1;
		if (mb[0]) io_pulse_1 <= 1;
		if (mb[1]) io_pulse_2 <= 1;
		if (mb[2]) io_pulse_4 <= 1;

		if (io_select == 6'b0)
		  begin
//temp-debug
//		    if (mb[2:0] == 3'b001)
//			interrupt_enable <= 1;
		    if (mb[2:0] == 3'b010)
			interrupt_enable <= 0;
		  end

		if (mb[0])
		  begin
		    case (io_select)
//		     6'o03: if (rx_int) pc <= pc + 1;
		     6'o04:
			begin
$display("tsf; tx_int %b, pc %o", tx_int, pc);
//			  if (tx_int) pc <= pc + 1;
//$display("after pc %o", pc);
			end
                    endcase
		  end

		if (mb[1])
		  begin
		    case (io_select)
		     6'o03: rx_int <= 0;
		     6'o04: tx_int <= 0;
                    endcase
		  end

		if (mb[2])
		  begin
		    case (io_select)
		     6'o03: ac <= rx_data;
		     6'o04:
			  begin
$display("tls; %o", ac);
			    tx_data <= ac;
			    tx_int <= 1;
			    tx_delaying <= 1;
			    tx_delay <= 4'b1111;
//interrupt <= 1;
$display("set tx_int");
			  end
                    endcase
		  end

//		if (io_select == 6'o4)
//		  begin
//			interrupt <= 1;
//		  end


		//??
/*
		if (io_pulse_1)
		  if (io_skip_flag[io_select])
		    pc <= pc + 1;

		if (io_pulse_4)
		  output_register[io_select] <= ac;

		if (io_pulse_2)
		  ac <= input_register[io_select];
*/
	  end

//	if (!(opr || iot) && !interrupt_cycle)
//	  pc <= pc + 1;

//$display("skip %b, incr %b", pc_skip, pc_incr);
	if (pc_skip)
	  pc <= pc + 2;
	else
	  if (pc_incr)
	    pc <= pc + 1;

	next_state <= F2;
	end

	F2:
	  begin
	    if (opr)
		ma <= pc;

	    if (iot)
		ma <= pc;

	    if (!(opr || iot))
	      begin
		ma[6:0] <= mb[6:0];
		if (!mb[7])
		  ma[11:7] <= 0;
	      end

	  next_state <= F3;
	  end

	F3:
	  begin
	    if (opr)
	      begin
		// group 1
		if (!mb[8])
		  begin
		    if (mb[0])
		      {l,ac} <= {l,ac} + 1'b1;
		    if (mb[3] && !mb[1])
		      {l,ac} <= {l,ac} >> 1;
		    if (mb[3] && mb[1])
		      {l,ac} <= {l,ac} >> 2;
		    if (mb[2] && !mb[1])
		      {l,ac} <= {l,ac} << 1;
		    if (mb[2] && mb[1])
		      {l,ac} <= {l,ac} << 2;
		  end

		if (mb[8] & !mb[0])
		  if (mb[2])
		     ac <= ac | switches;
		  if (mb[1])
		     run <= 0;

		ir <= 0;
		mb <= 0;
		next_state <= F0;
	      end

	    if (iot)
	      begin
		ir <= 0;
		mb <= 0;
		next_state <= F0;
	      end

	    if (!(opr || iot))
	      begin
		if (!mb[8] & jmp)
		  begin
		    pc <= ma;
		    ir <= 0;
		    mb <= 0;
		    next_state <= F0;
		  end
		
		if (mb[8])
		  begin
		    mb <= 0;
		    next_state <= D0; /* defer */
		  end

		if (!mb[8] & !jmp)
		  begin
		    mb <= 0;
		    next_state <= E0;
		  end
	      end
	end

      // DEFER

      D0:
	begin
	  mb <= memory_bus;
	  next_state <= D1;
	end

      D1:
	begin
	  // auto increment regs
	  if (ma[11:3] == 8'h01)
	    mb <= mb + 1;

	  next_state <= D2;
	end

      D2:
	begin
	  // write back
	  if (ma[11:3] == 8'h01)
	    ram_we_n <= 0;

	  ma <= #1 mb;
	  next_state <= D3;
	end

      D3:
	begin
	  ram_we_n <= 1;

	  if (jmp)
	    begin
	      pc <= mb;
	      ir <= 0;
	      mb <= 0;
	      next_state <= F0;
	    end

	  if (!jmp)
	    begin
	      mb <= 0;
	      next_state <= E0;
	    end
	end

      // EXECUTE

      E0:
	begin
	  mb <= memory_bus;
	  next_state <= E1;
	end

      E1:
	begin
	  if (and)
	    ;

	  if (isz)
	    begin
	      mb <= mb + 1;
	      if (mb == 12'b0)
	        pc <= pc + 1;
	    end

	  if (dca)
	    mb <= ac;

	  if (jms)
	    mb <= pc;

	  next_state <= E2;
	end

      E2:
	begin
	  if (isz || dca || jms)
begin
//$display("write %o %o", ma, mb);
	    ram_we_n <= 0;
end

	  if (~jms)
	    ma <= #1 pc;

	  if (jms)
	    ma <= #1 ma + 1;

	  next_state <= E3;
	end

      E3:
	begin
	  ram_we_n <= 1;

	  if (and)
	    ac <= ac & mb;

	  if (tad)
	    {l,ac} <= ac + mb;

	  if (dca)
	    ac <= 0;

	  if (jms)
	    pc <= ma;

	  ir <= 0;
//	  mb <= 0;
	  next_state <= F0;
	end
    endcase
  end

endmodule

/* 4kx12 static ram */
module ram_4kx12(A, DI, DO, CE_N, WE_N);

  input[11:0] A;
  input[11:0] DI;
  input CE_N, WE_N;
  output[11:0] DO;

  reg[11:0] ram [0:4095];
  integer i;

  initial
    begin
      for (i = 0; i < 4096; i=i+1)
        ram[i] = 12'b0;

	ram[12'o0000] = 12'o5177;
	ram[12'o0200] = 12'o7300;
	ram[12'o0201] = 12'o1300;
	ram[12'o0202] = 12'o1301;
	ram[12'o0203] = 12'o3302;
	ram[12'o0204] = 12'o7402;
	ram[12'o0205] = 12'o5200;

`include "focal.v"
	ram[12'o0000] = 12'o5404;
	ram[12'o0004] = 12'o0200;
    end

//  always @(posedge WE_N)
  always @(negedge WE_N)
    begin
      if (CE_N == 0)
        begin
$display("ram: write [%o] <- %o", A, DI);
          ram[ A ] = DI;
        end
    end

//always @(A)
//  begin
//    $display("ram: ce %b, we %b [%o] -> %o", CE_N, WE_N, A, ram[A]);
//  end

//  assign DO = ram[ A ];
assign DO = (^A === 1'bX || A === 1'bz) ? ram[0] : ram[A];

endmodule

